import os
import math
import sys

import fileheader

# This function estimates the number of clock cycles
# required for a particular adder chain (ac). 
# the second parameter (us) is the number of square
# block that is cascaded back to back.
def estimate_cycles(ac, us, ffield):
	l = len(ac) - 2                     # len(ac)-2 for quad, len(ac)-1 for sqr
	mul = 0
	sq = 0
	for i in range(1, l):
		p1 = int(ac[i-1])               # p1 + p2 = p, p2 >= p1, p2 = ac[i-1]
		p2 = int(ac[i]) - p1
		if(p2 <= us):
			sqs =  1
		else:
			sqs =  math.ceil(float(p2)/float(us))
		mul = mul + 1
		sq = sq + sqs
		print ac[i], p2, sqs, mul, sq+mul
	sq = sq + 1                         # the final squaring
	sq = sq + 2                         # add 2, one for loading, one for result
	print "squarings = ", sq, " multiplications = ", mul, "Total Clocks ", mul+sq

	fd = open("output/cc"+str(ffield), "w")
	fd.write(str(mul+sq))
	fd.close()
	return int(mul+sq)


def mulclk(fp, ss, ps, sel1, sel2, state):
	fp.write("\t"+str(ss)+"'d"+str(state)+":begin\n")
	fp.write("\t\tsel1 <= 2'd"+sel1+";")
	fp.write(" sel2 <= 2'd"+sel2+";")
	fp.write(" sel3 <= 1'bx;\n")
	fp.write("\t\ten   <= 1'b1;\n")
	fp.write("\t\tpow <= "+str(ps)+"'bx;\n")
	fp.write("\tend\n")

def powclk(fp, ss, ps, sel3, pow, state):
	fp.write("\t"+str(ss)+"'d"+str(state)+":begin\n")
	fp.write("\t\tsel1 <= 2'bx;")
	fp.write(" sel2 <= 2'bx;")
	fp.write(" sel3 <= 1'd"+sel3+";\n")
	fp.write("\t\ten   <= 1'b0;\n")
	fp.write("\t\tpow <= "+str(ps)+"'d"+pow+";\n")
	fp.write("\tend\n")


# Call this to define output states reqiuring
# more than one power operations
def multipow(fp, ss, ps, p2, qlen, state):
	n = p2 / qlen
	r = p2 % qlen
	
	powclk(fp, ss, ps, "0", str(qlen), state)
	state = state + 1
	for i in range(1, n):
		powclk(fp, ss, ps, "1", str(qlen), state)
		state = state + 1
	
	if r != 0:
		powclk(fp, ss, ps, "1", str(r), state)
		state = state + 1
	return state
	
# ac is the adder chain
# ffield is the finite field 
# qlen is the number of cascaded power blocks
def make_hdl(ac, ffield, qlen):
	fp = open("hdl/itoh.v", "w")

	fileheader.make(fp, "itoh.v", "Control Block for Itoh Tsujii", "control.py")
	fp.write("`ifndef __ITOH_V__\n")
	fp.write("`define __ITOH_V__\n")
	fp.write("`include \"quadblk.v\"\n")
	fp.write("`include \"mux4.v\"\n")
	fp.write("`include \"mux2.v\"\n")
	fp.write("`include \"multiplier.v\"\n\n")

	fp.write("module itoh_tsujii(clk, nrst, a, d);\n")
	fp.write("input wire clk;\n")
	fp.write("input wire nrst;\n")
	fp.write("input wire ["+str(ffield-1)+":0] a;\n")
	fp.write("output wire ["+str(ffield-1)+":0] d;\n\n")

	fp.write("wire ["+str(ffield-1)+":0] pin;\n")
	fp.write("wire ["+str(ffield-1)+":0] pout;\n")
	fp.write("wire ["+str(ffield-1)+":0] min1;\n")
	fp.write("wire ["+str(ffield-1)+":0] min2;\n")
	fp.write("wire ["+str(ffield-1)+":0] mout;\n")

	# ps is the number of bits required to store
	# the maximum power possible with the quadblock
	ps = int(math.ceil(math.log(qlen)/math.log(2)))
	fp.write("reg ["+str(ps-1)+":0] pow;\n")
	cc = estimate_cycles(ac, qlen, ffield)
	# ss is the number of bits required to store the states
	ss = int(math.ceil(math.log(cc)/math.log(2)))
	fp.write("reg ["+str(ss-1)+":0] state;\n\n")
	fp.write("reg ["+str(ss-1)+":0] nextstate;\n\n")

	# Define the multiplexer select lines and the buffer enable
	fp.write("reg [1:0] sel1;\n");
	fp.write("reg [1:0] sel2;\n");
	fp.write("reg       sel3;\n");
	fp.write("reg       en;\n");
	
	fp.write("wire      a1_en;\n");

	# We assume the chain is perfect, so no temporary
	# registers are required accept one.
	# instead there is only doublings or multiplications with input
	fp.write("reg ["+str(ffield-1)+":0] a1, a2;\n\n")

	# r1 and r2 are flipflops which latches the output of the multiplier
	# and power circuit.
	fp.write("reg ["+str(ffield-1)+":0] r1, r2;\n\n")

	# Generate the instances
	fp.write("pow4blk     pow4(pin, pow, pout);\n")
	fp.write("multiplier  mul(min1, min2, mout);\n\n")
	fp.write("mux4        m1(a, a1, r1, "+str(ffield)+"'bx, sel1, min1);\n") 
	fp.write("mux4        m2(a, r2, r1, "+str(ffield)+"'bx, sel2, min2);\n")
	fp.write("mux2        m3(r1, r2, sel3, pin);\n\n")
	
	fp.write("assign d = r1;\n")
	# The FSM
	fp.write("always @(posedge clk or negedge nrst) begin\n")
	fp.write("\tif (nrst == 1'b0)\n")
	fp.write("\t\tstate <= "+str(ss)+"'d0;\n")
	fp.write("\telse\n")
	fp.write("\t\tstate <= nextstate;\n")
	fp.write("end\n\n")

	# D flipflops to latch outputs of multiplier or quadblock
	fp.write("always @(posedge clk) begin\n")
	fp.write("\tif (en)\n")
	fp.write("\t\tr1<= mout;\n")
	fp.write("\telse\n")
	fp.write("\t\tr2<= pout;\n")
	fp.write("end\n\n")

	# D Flipflop to latch output of a1^3
	fp.write("assign a1_en = (state == "+str(ss)+"'d2) ? 1 : 0;\n");
	fp.write("always @(posedge a1_en) begin\n")
	fp.write("\ta1 <= mout;\n")
	fp.write("\ta2 <= mout;\n");
	fp.write("end\n\n")

	

	# Output Logic
	fp.write("always @(state) begin\n")
	fp.write("\tcase(state)\n")
	
	state = 0;
	mulclk(fp, ss, ps, "0", "0", state)      # a^2
	
	state = state + 1
	mulclk(fp, ss, ps, "0", "2", state)      # a^2.a = a1
	state = state + 1
	fp.write("\t/********** 2 p1:1 p2:1 *********/\n");
	powclk(fp, ss, ps, "0", "1", state)  # a1^(4^1) 
	state = state + 1
	mulclk(fp, ss, ps, "2", "1", state)      # a1^(4^1).a1 = a2
	state = state + 1

	for i in range(2, len(ac)-2):	   # for quad
#	for i in range(2, len(ac)-1):      # for square
		p1 = int(ac[i-1])
		p2 = int(ac[i]) - p1
		fp.write("\t/********** "+ str(ac[i])+" p1:"+str(p1)+" p2:"+str(p2)+" *********/\n");
	
		if p1 != p2 and p2 != 1:
			print "Unsupported Addition Chain"
			print ac, p1, ac[i]
			os.exit()
			break
		if(p2 <= qlen):        # If we need only one squaring
			powclk(fp, ss, ps, "0", str(p2), state)
			state = state + 1
		else:                  # If we need several squaring
			state = multipow(fp, ss, ps, p2, qlen, state)

			
		if (p1 == p2):         # doubling
			mulclk(fp, ss, ps, "2", "1", state)
		else:                  # not doubling then multiply with a1
			mulclk(fp, ss, ps, "1", "1", state)
		state = state + 1
		
	# The final squaring
	fp.write("\t/********* The Final Squaring *********/\n")
	mulclk(fp, ss, ps, "2", "2", state)
	state = state + 1 

	fp.write("\t"+str(ss)+"'d"+str(state)+":begin\n")
	fp.write("\t\tsel1 <= 2'bx; sel2 <= 2'bx; sel3 <= 1'bx;\n")
	fp.write("\t\tpow <= 4'bx;\n")
	fp.write("\t\ten <= 1'b0;\n")
	fp.write("\tend\n")

	# Default state
	fp.write("\tdefault: begin\n");
	fp.write("\t\tsel1 <= 2'bx; sel2 <= 2'bx; sel3 <= 1'bx;\n")
	fp.write("\t\tpow <= 4'bx;\n")
	fp.write("\t\ten <= 1'bx;\n")
	fp.write("\tend\n")

	fp.write("endcase\n")
	fp.write("end\n\n")

	# Now, do the next state logic
	fp.write("always @(state) begin\n")
	fp.write("\tcase (state)\n")
	for i in range(0, state):
		fp.write("\t"+str(ss)+"'d"+str(i)+": nextstate <= "+str(ss)+"'d"+str(i+1)+";\n")
	fp.write("\t"+str(ss)+"'d"+str(state)+": nextstate <= "+str(ss)+"'d"+str(state)+";\n")
	fp.write("\tdefault: nextstate <= "+str(ss)+"'bx;\n")
	fp.write("\tendcase\n")
	fp.write("end\n")




	fp.write("endmodule\n")
	fp.write("`endif\n")

	prj = open("hdl/itoh.prj", "a")
	prj.write("verilog work itoh.v\n")
	prj.close()
	


	fp.close()
