#include <stdio.h>

#include "makeheader.h"


#define CAPACITY 1000
// Irreducible Polynomial x^233 + x^74 + 1 = 0
#define FFIELD  233
#define TRIN 74     



/* Each bit of the result is expressed as xor of several other
 * bits. xormap contains the bits which on adding will give result.
 * ie. result[i] = bit_xormap[0] ^ bit_xormap[i] ....
*/
struct result{
	unsigned int xormap[40000]; 
	int s;	
} res[CAPACITY];


unsigned int ffield;
unsigned int trin;


void print()
{
	int i,j;
	for(i=0; i<CAPACITY; ++i){
		if(res[i].s != 0){
			printf("%03d : ", i);
			for(j=0; j<res[i].s; ++j)
				if(res[i].xormap[j] != ~0x0)	printf("%03d ", res[i].xormap[j]);
			printf("\n");
		}
	}
}

void make_verilog(char *name, char *ifdefnam)
{
	FILE *fd;
	char buf[50];
	int i, j;
	int xorcount, xors;

	sprintf(buf, "hdl/%s.v", name);
	fd = fopen(buf, "w");

	makeheader(fd, buf, "Modulo Operation", "modverilog.c");
	fprintf(fd, "`ifndef %s\n", ifdefnam);
	fprintf(fd, "`define %s\n", ifdefnam);

	fprintf(fd, "module %s(a, d);\n\n", name);
	fprintf(fd, "input wire [%d:0] a;\n", ffield*2-2);
	fprintf(fd, "output wire [%d:0] d;\n\n", ffield-1);

	for(i=0, xorcount=0; i<ffield; ++i){
		fprintf(fd, "assign d[%d] = ", i);
		if(res[i].s == 0){
			fprintf(fd, " 0;\n");
		}else if(res[i].s != 0){
			for(j=0, xors=0; j<res[i].s; ++j)
				if(res[i].xormap[j] != ~0x0){
					fprintf(fd, "a[%d] ^ ", res[i].xormap[j]);
					xorcount++;
					xors++;
			}
			fseek(fd, -3, SEEK_CUR);  // remove the last ^
			if(xors)
				xorcount--;           // remove the last xorcount
			else
				fprintf(fd, "=0");
			fprintf(fd, ";\n");
		}
	}
	fprintf(fd, "endmodule\n");
	fprintf(fd, "`endif\n");
	close(fd);
	printf("Xor count in file %s = %d\n", name, xorcount);
}

void insert(unsigned int r, unsigned int val)
{
	int i,j;
	int len = res[r].s;
	for(i=0; i<len; ++i){
		if(res[r].xormap[i] == val){
			/* Delete it */
			for(j=i+1; j<len; ++j){
				res[r].xormap[j-1] = res[r].xormap[j];
			}
			res[r].xormap[len-1] = 0;
			res[r].s--;
			return;
		}
	}
	res[r].xormap[len] = val;
	res[r].s++;
	return;
}

/* reduce: do the modulo reduction
 * 
*/
int reduce(void)
{
	int i,j;
	for(i=ffield; i<CAPACITY; ++i){
		if(res[i].s == 0) 
			continue;
		for(j=0; j<res[i].s; ++j){
			insert(i-ffield, res[i].xormap[j]);
			insert(i+trin-ffield, res[i].xormap[j]);
//			res[i-FFIELD].xormap[res[i-FFIELD].s++] 
//				= res[i].xormap[j];
//			res[i+TRIN-FFIELD].xormap[res[i+TRIN-FFIELD].s++]
//				= res[i].xormap[j];
			res[i].xormap[j] = 0;
			
		}
		res[i].s = 0;
				
	}
}
/* square the polynomial
*/
void square()
{
	int i, j;
	for(i=CAPACITY/2 -1; i>0; --i){
		for(j=0; j<res[i].s; ++j){
			res[i*2].xormap[j] = res[i].xormap[j];
			res[i].xormap[j] = 0;
		}
		res[i*2].s = res[i].s;
		res[i].s = 0;
	}
}

/*
 * x ^ x = 0, eliminates such terms
*/ 
void trim()
{
	int i, j, k;
	int size;
	for(i=0; i<CAPACITY; ++i){
		size = res[i].s;
		for(j=0; j<size - 1; ++j){
			for(k=j+1; k<size; ++k){
				if(res[i].xormap[j] == res[i].xormap[k]){
					res[i].xormap[k] = ~0x0;
					res[i].xormap[j] = ~0x0;
//					res[i].s -= 2;
					break;
				}
			}		
		}
	}	
}

/* generate verilog code for the polynomial */

int main(int argc, char **argv){
	int i,j;
	FILE *fd;
	
	ffield = FFIELD;
	trin = TRIN;
	//printf("%d\n", atoi(argv[1]));
	if(argc >= 2)
		ffield = atoi(argv[1]);
	if(argc >= 3)
		trin = atoi(argv[2]);
	printf("Generating modulo for IR Poly {%d %d 1}\n", ffield, trin);
	

		
	/* Initialize */
	memset(res, 0x0, sizeof(struct result)* CAPACITY);
	for(i=0; i<(ffield*2)-1; ++i){
		res[i].xormap[res[i].s] = i;
		res[i].s++;
	}

	reduce(); reduce(); 
	trim(); 
	make_verilog("mod", "__MOD_V__");

	fd = fopen("hdl/itoh.prj", "a");
	fprintf(fd, "verilog work mod.v\n");
	close(fd);
}
